/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * console window functions
 */

#include <stdio.h>

#include "console.h"

#include "text.h"
#include "rli.h"
#include "common.h"
#include "screen.h"
#include "menus.h"
#include "and_scr.h"
#include "edit.h"
#include "windows.h"
#include "and_file.h"
#include "ezconfig.h"

#define CON_BUFFER_SIZE	4096

extern TEXT_VIEW *text_list;

TEXT_VIEW *cons_textview;

void CON_SetupConsole(void)
{
    config_position_t pos;
  
    config_getpos("ConsoleWindow",&pos,0,1,80,5);
  
    cons_textview = TXT_initialize(pos.y, pos.x,
				   pos.dimy, pos.dimx,
				   (long)CON_BUFFER_SIZE,
				   (TF_TRIM | TF_SCROLL | TF_EXACT));
    if (cons_textview == (TEXT_VIEW *)NULL) {
	SCR_RestoreVideoEnv();
	CON_WriteLine(MSG_NotEnoughMemory);
	exit(-1);
    }
}

void CON_SetNoConnBanner(void)
{
    TXT_SetBanner(cons_textview->a_tbuf, con_noconn_banner);
    CON_SetBusy(FALSE);
}

void CON_SetBanner(char *aBan)
{
  TXT_SetBanner(cons_textview->a_tbuf, aBan);
}

void CON_Write(char *str)
{
  static int need_cr = FALSE;

  if (cons_textview) {
    int len = strlen(str);
    int need_before_cr = need_cr;
    need_cr = FALSE;
    while (*str == '\n') {
      str++; len--;
      need_before_cr = TRUE;
    }
    while (len > 0 && str[len-1] == '\n') {
      len--;
      need_cr = TRUE;
    }
    if (len > 0) {
      if (need_before_cr)
	TXT_append(cons_textview->a_tbuf, "\n", 1);
      TXT_append(cons_textview->a_tbuf, str, len);
    } else
      need_cr = (need_cr || need_before_cr ? TRUE : FALSE);
  } else
    printf("%s", str);
}

void CON_Show(void)
{
  WINLIST *awp = cons_textview->a_tbuf->awp;

  WIN_Show(awp);
  WIN_BringToFront(awp);
}

void CON_SetupMenus(int net_state)
{
  if (WIN_FrontWindow(cons_textview->a_tbuf->awp)) {
    MNU_EnableItem(M_CLOSE, FALSE);
  }
}

int CON_IsFrontWindow(void)
{
  return(WIN_FrontWindow(cons_textview->a_tbuf->awp));
}

void CON_ReportSize(void)
{
  WINFO *w = cons_textview->a_tbuf->awp->awin;

  /*xxx  a_local_prefs->loc[CONSOLE_RL].y = w->loc.y;
  a_local_prefs->loc[CONSOLE_RL].x = w->loc.x;
  a_local_prefs->dim[CONSOLE_RL].y = w->dim.y;
  a_local_prefs->dim[CONSOLE_RL].x = w->dim.x; */
}

void CON_ChangedSize(void)
{
  WINFO *w = cons_textview->a_tbuf->awp->awin;

  /* w->loc.y = a_local_prefs->loc[CONSOLE_RL].y;
  w->loc.x = a_local_prefs->loc[CONSOLE_RL].x;
  w->dim.y = a_local_prefs->dim[CONSOLE_RL].y;
  w->dim.x = a_local_prefs->dim[CONSOLE_RL].x;
  */
  WIN_WindowChangedSize(w);
  TXT_Draw(cons_textview->a_tbuf->awp);
}

void CON_WriteLine(char *str)
{
  CON_Write(str);
  CON_Write("\n");
}

int CON_CanClose(void)
{
  /*
    this is called at quit time.  Try to close all documents.
    if all closed return Ok. Otherwise, return false.
  */

  TEXT_VIEW *atl;

  for (atl = text_list; atl != NULL; atl = atl->next) {
    TEXT_BUF *tbuf = atl->a_tbuf;
    if (atl == cons_textview) continue;
    if (TF_ISSET(tbuf->flags, TF_CHANGED)) {
      WIN_BringToFront(tbuf->awp);
      if (!TXT_Close(tbuf->awp)) return(FALSE);
    } 
    else {
      if (TF_ISSET(tbuf->flags, TF_DOCUMENT)) {
	TXT_Close(tbuf->awp);
      }
    }
  }
  return(TRUE);
}

char *CON_GetHandle(void)
{
  return(cons_textview->a_tbuf->buf);
}

int CON_GetHandleSize(void)
{
  cons_textview->a_tbuf->buf[cons_textview->a_tbuf->fsize] = 0;
  return((int)cons_textview->a_tbuf->fsize);
}

#define      busy_indicator_off   NIL
static char *busy_indicator_on = "BUSY";

extern Boolean g_while_list_printing;

void CON_SetBusy(int on_off)
{
  gWeAreBusy = (Boolean)on_off;

  on_off = (g_while_list_printing ? TRUE : gWeAreBusy);

  SCR_Write((on_off ? busy_indicator_on : busy_indicator_off),
	    0, imax(0, (screen_cols - 5)), 4, (on_off ? RVIDEO : NORMAL));

  if (!gNovice && gWeAreBusy == FALSE) SCR_ClearStatusLine();

  SCR_UpdateScreen();
}

void CON_RedrawBusyState(void)
{
  CON_SetBusy(gWeAreBusy);
}

#ifndef IBMPC

static Boolean the_end_is_here = FALSE;
static Boolean already_saving_it = 0;

void save_the_world(void)
{
  the_end_is_here = TRUE;
  if (already_saving_it) return;
  already_saving_it++;
  checkpoint_openfiles();
}

void checkpoint_openfiles(void)
{
  extern char *FIO_GenUniqueName(), *getenv();
  static int count = 1;
  TEXT_VIEW_pt atview;

  for (atview=text_list; atview; atview = atview->next) {
    if (atview == cons_textview) continue;
    if (!TF_ISSET(atview->a_tbuf->flags, TF_DOCUMENT)) continue;
    if (TF_ISSET(atview->a_tbuf->flags, TF_OFFDSK)) continue;
    if (!TF_ISSET(atview->a_tbuf->flags, TF_CHANGED) &&
	!atview->a_tbuf->modify_count) continue;
    if (atview->a_tbuf->ckpt_count || the_end_is_here) {
      TEXT_BUF_pt atbuf = atview->a_tbuf;
      char path[MAXPATHLEN], fullpath[MAXPATHLEN];
      EDT_CloseGap(atbuf); /* close gap if editing */
      if (the_end_is_here) {
	sprintf(path, "~/Mailbox/%s%d.CKP", FIO_GenUniqueName(), count++);
	if (FIO_ResolveName(path, fullpath) == 0)
	  TXT_WriteFile(atbuf, fullpath);
      }
      if (!atbuf->ckpt_file[0]) {
	if (FIO_GenTmpName(path) == 0)
	  sprintf(fullpath, "%s%d.CKP", path, count++);
	if (FIO_ResolveName(fullpath, atbuf->ckpt_file))
	  sprintf(atbuf->ckpt_file,"/tmp/%s%d.CKP",FIO_GenUniqueName(),count);
      }
      TXT_WriteFile(atbuf, atbuf->ckpt_file);
      atbuf->ckpt_count = 0;
    }
  }
}
#endif /* not IBMPC */
