/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * document handling routines
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>

#include <rli.h>
#include <common.h>
#include <screen.h>
#include <comkeys.h>
#include <menus.h>
#include "xmalloc.h"
#include "common.h"
#include "document.h"
#include "smtp.h"
#include "and_file.h"
#include "console.h"
#include "edit.h"
#include "dialog.h"
#include "windows.h"
#include "clinet.h"
#include "edit.h"
#include "alias.h"
#include "rlci.h"
#include "text.h"
#include "ezconfig.h"

#define DOC_OP_LOCAL  0
#define DOC_OP_REMOTE 1

#define RW_BUFSIZE 2048

extern FILE *FIO_OpenStream();

extern char *get_machine_config(), *CON_GetHandle();

extern char *RFile_filename();
extern long  RFile_datasize();

#define DOC_F_EDIT       0x01
#define DOC_F_VERBOSE    0x02
#define DOC_F_REMOVE     0x04
#define DOC_F_EXPECT_ERR 0x08
#define DOC_F_EDITBODY   0x10

typedef struct doc_struct {
  int flags;                      /* Flags associated with this transcation */
  int menu_cmd;                   /* Menu that triggered the action */
  char *buf;                      /* Local temporary buffer */
  long  bufsize;                  /* Size of the allocated buffer */
  long  fsize;                    /* Real size of the file */
  long  offset;                   /* Current offset into the file */
  TEXT_VIEW_pt textview;          /* The view associated with this document */
  TEXT_VIEW_pt save_tview;        /* The view used for replace file */
  FILE  *stream;                  /* The stream if read/write from file */
  char  local_fname[MAXPATHLEN];  /* Filename associated with the stream */
    
    char *answer_mbox;
    int answer_num;
} DOCUMENT, *DOCUMENT_pt;

extern alias_t *g_aliases;

DOCUMENT_pt g_waiting_doc = (DOCUMENT_pt)NULL;
DOCUMENT    g_document;

/*
 * This module only works on one document at a time.
 */

static int DOC_IsBusy()
{
  /*
   * are we already busy with an existing operation?
   */

  if (g_waiting_doc) {
    DLG_Alert(ALERT_warning, MSG_FileIO_InProgress);
    return(TRUE);
  }
  return(FALSE);
}

static char *DOC_UntitledName()
{
  static int  untitled_count = 0;
  static char untitled_name[STR255LEN];

  sprintf(&untitled_name[0], "Untitled.%d", ++untitled_count);
  return(&untitled_name[0]);
}

TEXT_VIEW_pt DOC_CreateTextview(long file_size)
{
  TEXT_VIEW_pt atview;
  config_position_t pos;
  
  config_getpos("SendWindow",&pos,0, 6, 80, screen_rows-7);
  
  atview = TXT_initialize(pos.y, pos.x,
			  pos.dimy, pos.dimx,
			  (long)(file_size),
			  (TF_CLOSE | TF_FREE | TF_DOCUMENT | TF_LOCKED)
			  );

  if (!atview) CON_WriteLine(MSG_NotEnoughMemory);

  return(atview);
}

void DOC_ActivateView(TEXT_BUF_pt atbuf)
{
    const char *editor = config_getstring("ExternalEditor",NULL);

  WIN_Show(atbuf->awp);
  WIN_BringToFront(atbuf->awp);

  if (TF_ISSET(atbuf->flags, TF_EDIT)) {
    TF_CLEAR(atbuf->flags, TF_EDIT);
    if ((editor!=NULL) && (strlen(editor)>0)) {
	DoMenuCommand(M_EXTERNAL_EDIT);
    } else {
	DoMenuCommand(M_EDITMSG);
	if (g_waiting_doc && TF_ISSET(g_waiting_doc->flags, DOC_F_EDITBODY))
	    EDT_SetCursorInBody(atbuf);
    }
  }
  TF_CLEAR(atbuf->flags, TF_LOCKED);
}

static int DOC_AllocateViewBuffer(aDoc, size)
DOCUMENT_pt aDoc;
long size;
{
    const char *tmpdir = config_getstring("TempDir",NULL);

  if (size >= PAGEABLE_BUFSIZE && tmpdir) {
    if (FIO_GenTmpName(aDoc->local_fname))
      DLG_Alert(ALERT_warning, MSG_TempError);
    else {
      aDoc->stream = FIO_OpenStream(aDoc->local_fname, OPEN_WT);
      if (aDoc->stream == (FILE *)NULL)
	DLG_Alert(ALERT_warning, MSG_OpenFailed);
    }
    if (aDoc->stream == (FILE *)NULL) return(-1);
  }
  return(0);
}

static int DOC_AllocRWBuffer(aDoc)
DOCUMENT_pt aDoc;
{
  aDoc->buf = (char *)malloc((unsigned int)(RW_BUFSIZE + 1));

  if (aDoc->buf == NIL) {
    DLG_Alert(ALERT_warning, MSG_NotEnoughMemory);
    return(-1);
  }

  aDoc->bufsize = RW_BUFSIZE;
  return(0);
}

static void DOC_SetWindowTitle(aDoc, menu_cmd, name, banner)
DOCUMENT_pt aDoc;
int menu_cmd;
char *name, *banner;
{
  TEXT_BUF_pt atbuf = aDoc->textview->a_tbuf;

  if (menu_cmd==M_QUICK_EDIT || menu_cmd==M_AFS_EDIT || menu_cmd==M_AFS_NEW) {
    TF_SET_AFSFILE(atbuf->flags);
    WIN_SetTitle(atbuf->awp, name);
  } else
    WIN_SetTitle(atbuf->awp, DOC_UntitledName());

  if (banner) TXT_SetBanner(atbuf, banner);
}

static int DOC_Initialize(op_type, name, size)
int op_type;
char *name;
long size;
{
  int save_flags = g_document.flags;
  int menu_cmd   = g_document.menu_cmd;

  bzero(&g_document, sizeof(DOCUMENT));

  if (DOC_AllocateViewBuffer(&g_document, (long)size)) {
    if (size >= PAGEABLE_BUFSIZE && g_document.stream == (FILE *)NULL)
      DLG_Alert(ALERT_warning, MSG_GetFile);
    return(-1);
  }

  if (op_type == DOC_OP_LOCAL || DOC_AllocRWBuffer(&g_document) == 0)
    g_document.textview = DOC_CreateTextview((long)size);

  if (g_document.textview == (TEXT_VIEW *)NULL) {
    if (g_document.buf) free(g_document.buf);
    if (g_document.stream) {
      FIO_CloseStream(g_document.stream);
      FIO_EraseFile(g_document.local_fname);
    }
    bzero(&g_document, sizeof(DOCUMENT));
    return(-1);
  }

  g_document.fsize = size;

  if (TF_ISSET(save_flags, DOC_F_EDIT))
    TF_SET_EDIT(g_document.textview->a_tbuf->flags);

  g_document.flags = save_flags & (DOC_F_EDITBODY | DOC_F_REMOVE);

  TF_SET_REMOVE(g_document.textview->a_tbuf->flags);

  strcpy(g_document.textview->a_tbuf->lname, g_document.local_fname);
  DOC_SetWindowTitle(&g_document, menu_cmd, name, doc_standard_banner);

  if (op_type == DOC_OP_REMOTE) {
    RFile_set_buf(g_document.buf, (long)g_document.bufsize);
    RFile_read((long)g_document.bufsize);
    g_waiting_doc = &g_document;
  }
  return(0);
}

static int DOC_WriteToView(DOCUMENT_pt aDoc,char *buf,int   len)
{
  int rc;

  if (aDoc->stream)
    rc = FIO_WriteStream(aDoc->stream, buf, len);
  else
    rc = TXT_append(aDoc->textview->a_tbuf, buf, len);

  return(rc);
}

void DOC_OpenLocalFile(void)
{
  long fsize;
  char filename[STR255LEN];
  const char *tmpdir = config_getstring("TempDir",NULL);

  filename[0] = '\0';
  if (get_local_string(MSG_OPEN_LOCAL, filename) && filename[0]) {
    char realname[MAXPATHLEN];
    if (FIO_ResolveName(filename, realname) == 0 && 
	(fsize = FIO_GetFileSize(realname)) >= 0) {
      if (fsize >= PAGEABLE_BUFSIZE && tmpdir) {
	char tempname[MAXPATHLEN];
	if (FIO_GenTmpName(&tempname[0]))
	  DLG_Alert(ALERT_warning, MSG_TempError);
	else {
	  if (FIO_CloneFile(&realname[0], &tempname[0], (long)fsize))
	    DLG_Alert(ALERT_warning, MSG_WriteError);
	  else
	    DOC_LoadLocalFile(&tempname[0], (long)fsize,
			      &filename[0], TRUE, TRUE);
	}
      } else
	  DOC_LoadLocalFile(&realname[0], (long)fsize,
			    &filename[0], FALSE, FALSE);
    } else
      CON_WriteLine(MSG_OpenFailed);
  }
}

int DOC_ReadFile(TEXT_BUF_pt atbuf,char *filename,long fsize)
{
  int rc;

  if ((rc = FIO_OpenFile(filename)) == 0) {
    rc = imin(fsize, atbuf->bsize);
    rc = FIO_ReadFromFile(&atbuf->buf[0], rc);
    if (rc > 0) atbuf->fsize = rc;
    FIO_CloseFile();
  }
  return(rc);
}

void DOC_LoadLocalFile(char *filename,long  fsize,char *window_name,
		       int remove_on_close, int off_disk)
{
  TEXT_VIEW *tview;
  tview = DOC_CreateTextview((long)fsize);
  if (tview) {
    TEXT_BUF *tbuf = tview->a_tbuf;
    if (off_disk) {
      tbuf->fsize = fsize;
      if (FIO_ResolveName(filename, tbuf->lname))
	strcpy(tbuf->lname, filename);
      TF_SET_FLOAD(tbuf->flags);
      TF_SET_OFFDSK(tbuf->flags);
      if (remove_on_close) TF_SET_REMOVE(tbuf->flags);
    } else {
      int rc = DOC_ReadFile(tbuf, filename, (long)fsize);
      if (rc < 0) DLG_Alert(ALERT_warning, MSG_ReadError);
    }
    WIN_SetTitle(tbuf->awp, (window_name ? window_name : filename));
    TXT_SetBanner(tbuf, doc_standard_banner);
    DOC_ActivateView(tbuf);
  }
}

void DOC_BugReport(void)
{
  int consize;
  int bufsize;

  if (DOC_IsBusy() == FALSE) {
    consize = CON_GetHandleSize();
    bufsize = strlen(bug_headers) + consize + 2048;

    TF_SET(g_document.flags, DOC_F_EDIT);
    if (DOC_Initialize(DOC_OP_LOCAL, NIL, (long)bufsize) == 0) {
      char bigbuf[2048], *ptr;
      DOC_WriteToView(&g_document, bug_headers, strlen(bug_headers));
      strcpy(&bigbuf[0], "Machine information:\n");
      ptr = get_machine_config(&bigbuf[strlen(&bigbuf[0])], 1000);
      strcpy(ptr, "Console log:\n");
      DOC_WriteToView(&g_document, bigbuf, strlen(bigbuf));
      DOC_WriteToView(&g_document, CON_GetHandle(), consize);
      DOC_ActivateView(g_document.textview->a_tbuf);
      bzero(&g_document, sizeof(DOCUMENT));
    }
  }
}

void DOC_OpenNew(void)
{
    /* check to make sure a return address is set */
    const char *email = config_getstring("Email",NULL);
    if ((email==NULL) || (strlen(email)==0)) {
	DLG_Alert(ALERT_warning,"No email addrees set. Set one in the preferences");
	return;
    }

  if (DOC_IsBusy() == FALSE) {
    int len = strlen(compose_headers);
    TF_SET(g_document.flags, DOC_F_EDIT);
    if (DOC_Initialize(DOC_OP_LOCAL, NIL, (long)len) == 0) {
      DOC_WriteToView(&g_document, compose_headers, len);
      DOC_ActivateView(g_document.textview->a_tbuf);
      bzero(&g_document, sizeof(DOCUMENT));
    }
  }

}

extern int cur_bodyline;

#define FORWARD_SEPERATOR "---Forwarded Message Follows---\n"

void DOC_OpenNew_withdata(char *to, char *cc, char *subject,
			  char *answer_mbox, int answer_num, int fwd)
{
    /* check to make sure a return address is set */
    const char *email = config_getstring("Email",NULL);
    if ((email==NULL) || (strlen(email)==0)) {
	DLG_Alert(ALERT_warning,"No email addrees set. Set one in the preferences");
	return;
    }
    


  if (DOC_IsBusy() == FALSE) {
      int len;
      char *str;
      TF_SET(g_document.flags, DOC_F_EDIT);

      if (!to) to = "";
      if (!cc) cc = "";
      if (!subject) subject="";

      len = strlen(compose_headers) + strlen(to) + strlen(cc) + strlen(subject);
      str = malloc(len+2);
      if (!str) {
	  DOC_OpenNew();
	  return;
      }
      
      snprintf(str,len+1,"To: %s\nSubject: %s\nCc: %s\n\n",
	       to, subject, cc);
      
      
      if (DOC_Initialize(DOC_OP_LOCAL, NIL, (long)len) == 0) {
	  DOC_WriteToView(&g_document, str, len);

	  if (fwd == 1) {
	      int i;
	      char line[100];
	      char *headers;

	      /* put in some blank lines */
	      DOC_WriteToView(&g_document, "\n",1);
	      DOC_WriteToView(&g_document, "\n",1);
	      DOC_WriteToView(&g_document, "\n",1);

	      
	      DOC_WriteToView(&g_document, FORWARD_SEPERATOR, strlen(FORWARD_SEPERATOR));

	      /* put in full headers */
	      headers = do_getheaders(do_getcurrentg(), answer_num);
	      DOC_WriteToView(&g_document, headers, strlen(headers));
	      
	      /* Put in full message body */
	      i = 0;
	      do {
		  i++;
		  RL_read(BODY_RL,i,line);
	      } while (line[0]!='\0');

	      for (;i<cur_bodyline;i++) {
		  
		  RL_read(BODY_RL,i,line);
		  strcat(line,"\n");
		  
		  DOC_WriteToView(&g_document, line, strlen(line));  
	      }
	  } else if (fwd == 2) {
	    /* JCS - revive the draft by reading the body 
	     * instead of doing fancy forwarding stuff */
	    int i;	    
	    char line[100];
	    
	    i = 0;
	    do {
	      i++;
	      RL_read(BODY_RL,i,line);
	    } while (line[0]!='\0');
    
	    for (;i<cur_bodyline;i++) {
	      
	      RL_read(BODY_RL,i,line);
	      strcat(line,"\n");
	      
	      DOC_WriteToView(&g_document, line, strlen(line));  
	    }	    
	  }

	  DOC_ActivateView(g_document.textview->a_tbuf);

	  EDT_SetCursorInBody(g_document.textview->a_tbuf);

	  if (!fwd) {
	      /* put in excerpt */
	      EDT_RestoreLine(g_document.textview->a_tbuf);
	  } else {
	      /* JCS - When forwarding, the first thing you want to 
	       * do is write the recipient's To: address, so reposition
	       * the cursor there instead.
	       */
	      EDT_SetCursorInHeader ( g_document.textview->a_tbuf );
	  }

	  bzero(&g_document, sizeof(DOCUMENT));	  
      }
      free(str);

      if (answer_mbox) {
	  g_document.answer_mbox = xstrdup(answer_mbox);
      } else {
	  g_document.answer_mbox = NULL;
      }
      g_document.answer_num = answer_num;

  }
}

int DOC_signmessage(TEXT_VIEW *tview)
{
    TEXT_BUF *tbuf = tview->a_tbuf;
    char *homedir = getenv("HOME");
    char filename[MAXPATHLEN+1];
    int size;
    FILE *stream;
    char *sig;
    
    if (!homedir) {
	DLG_Alert(ALERT_warning,"Can't find home directory");
	return -1;
    }
    
    snprintf(filename,sizeof(filename)-1,"%s/%s",homedir,".signature");

    size = FIO_GetFileSize(filename);    

    if (size <= 0) return -1;
    
    stream = fopen(filename,"r");
    
    sig = malloc(size+10);
    if (!sig) return -1;
    
    fread(sig,1,size,stream);
    sig[size] = '\0';

    fclose(stream);

    EDT_Edit(tbuf);
    tbuf->char_pos = tbuf->gap_start;

    EDT_InsertString(tbuf,"\n",1,0);
    EDT_InsertString(tbuf,sig,strlen(sig),0);

    EDT_EditEnd(tbuf);

    return 0;
}

void DOC_AppendLocal(TEXT_VIEW *tview,char *fname)
{
  int  rc;
  long fsize;
  TEXT_BUF *atbuf = tview->a_tbuf;
  char efname[MAXPATHLEN];
  const char *tmpdir = config_getstring("TempDir",NULL);

  if (FIO_ResolveName(fname, efname)||(fsize = FIO_GetFileSize(fname)) <= 0) {
    CON_WriteLine(MSG_OpenFailed);
    return;
  }

  if (TXT_ReallocTview(tview, (long)(atbuf->fsize + fsize + 1))) {
    CON_WriteLine(MSG_NotEnoughMemory);
    return;
  }

  if (!TF_ISSET(atbuf->flags, TF_OFFDSK) &&
      (atbuf->bsize < atbuf->fsize + fsize)) {
    if (tmpdir) {
      if (FIO_GenTmpName(atbuf->lname) == 0) {
	if (TXT_WriteFile(atbuf, atbuf->lname) == 0) {
	  TF_SET_OFFDSK(atbuf->flags);
	  TF_SET_REMOVE(atbuf->flags);
	}
      }
    }
  }

  if (TF_ISSET(atbuf->flags, TF_ISOPEN)) {
    FIO_CloseStream(atbuf->stream);
    TF_CLEAR(atbuf->flags, TF_ISOPEN);
  }

  if (TF_ISSET(atbuf->flags, TF_OFFDSK)) {
    if ((rc = FIO_AppendFile(efname, atbuf->lname, (long)fsize)) != 0)
      CON_WriteLine(MSG_WriteError);
    else
      atbuf->fsize = FIO_GetFileSize(atbuf->lname);
  } else {
    if ((rc = FIO_OpenFile(efname)) == 0) {
      char buf[1024];
      while (fsize > 0) {
	rc = FIO_ReadFromFile(&buf[0], sizeof(buf) - 1);
	if (rc <= 0) {
	  if (rc < 0) CON_WriteLine(MSG_ReadError);
	  break;
	}
	buf[rc] = '\0';
	if ((rc = TXT_append(atbuf, &buf[0], rc)) != 0) {
	  CON_WriteLine(MSG_AppendError);
	  break;
	}
	fsize -= rc;
      }
    }
  }

  if (rc == 0) CON_WriteLine(MSG_AppendDone);

  TF_SET_CHANGED(atbuf->flags);
  if (TF_ISSET(atbuf->flags, TF_OFFDSK)) TF_SET_FLOAD(atbuf->flags);
  TXT_Draw(atbuf->awp);
}

void DOC_AppendRemote(TEXT_VIEW *tview,char *fname)
{
  TEXT_BUF *tbuf = tview->a_tbuf;

  RFile_initialize();
  bzero(&g_document, sizeof(DOCUMENT));

  if (TF_ISSET(tbuf->flags, TF_OFFDSK)) {
    if (TF_ISSET(tbuf->flags, TF_ISOPEN)) {
      TF_CLEAR(tbuf->flags, TF_ISOPEN);
      FIO_CloseStream(tbuf->stream);
    }
    g_document.stream = FIO_OpenStream(tbuf->lname, OPEN_AT);
    if (g_document.stream == (FILE *)NULL) {
      DLG_Alert(ALERT_warning, MSG_AppendError);
      return;
    }
  }
  if (DOC_AllocRWBuffer(&g_document) == 0) {
    g_document.fsize = 0;
    g_document.textview = tview;
    g_document.menu_cmd = M_APND_REMOTE;
    TF_SET(g_document.flags, DOC_F_EXPECT_ERR);
    strcpy(g_document.local_fname, tbuf->lname);
    RFile_set_buf(g_document.buf, (long)g_document.bufsize);
    g_waiting_doc = &g_document;
  } else if (g_document.stream) FIO_CloseStream(g_document.stream);
}


static int verify_recipients(TEXT_BUF *tbuf, char *header)
{
    char **ret = xmalloc(sizeof(char *) * 20);
    char buf[2048];
    int bytes_read;
    int i;
    char *bufp;
    int size = 0;
    char *start = NULL;
    
    bytes_read = TXT_GetData(tbuf,
			     buf, sizeof(buf)-1, 0);
    bufp = buf;

    while (*bufp != '\0' ) {
	if (strncmp(bufp,header,strlen(header))==0) {
	    start = bufp + strlen(header);
	    break;
	}
	
	/* go to next line */
	bufp = strchr(bufp,'\n');
	if (!bufp) return -1;
	bufp++;
    }

    if (start == NULL) return -1;

    ret[0] = NULL;
    size = parse_recipients_line(start,ret,0,&bufp);
    
    EDT_Edit(tbuf);
    tbuf->char_pos = (start - buf)+1;

    for (i = 0;i< bufp - start;i++) {
	EDT_DeleteChar(tbuf);
    }

    /* now put the address text in */
    for (i = 0;i<size;i++) {

	if (ret[i] == NULL) continue;

	EDT_InsertString(tbuf,ret[i],strlen(ret[i]),0);

	if (i < size-1) {
	    EDT_InsertString(tbuf,", ",2,0);	
	}
    }
    
    EDT_EditEnd(tbuf);

    return 0;
}

void DOC_VerifyMsg(TEXT_VIEW *tview)
{
    int r;

    if (!DOC_SaveAsRemote(tview, NIL)) return;

    /* expand to and cc fields */
    r = verify_recipients(g_waiting_doc->textview->a_tbuf, "To:");
    if (r) {
	CON_WriteLine("Error expanding To: field");
    }
    r = verify_recipients(g_waiting_doc->textview->a_tbuf, "Cc:");
    if (r) {
	CON_WriteLine("Error expanding Cc: field");
    }

    TF_CLEAR(tview->a_tbuf->flags, TF_CHANGED);
    TF_CLEAR(tview->a_tbuf->flags, TF_LOCKED);

    /* JCS - this allows us to verify twice (or send the message after a 
     *  manual verify with the changes below!) 
     */
    g_waiting_doc = NULL;
}

void DOC_SubmitMsg(TEXT_VIEW *tview)
{
    smtp_t *smtp_conn;
    int r;
    const char *sentfolder = config_getstring("SentFolder",NULL);

    /* JCS - Sign the message */
    if ( config_getswitch("SignMail",FALSE) )
	DOC_signmessage(tview);
 
    /* JCS - redo the stuff we do in DOC_VerifyMsg(), but this 
     * time we actually send off the message and complete everything 
     *  for good 
     */
    if (!DOC_SaveAsRemote(tview, NIL)) return;
    
    TF_CLEAR(tview->a_tbuf->flags, TF_CHANGED);
    TF_CLEAR(tview->a_tbuf->flags, TF_LOCKED);
    
    /* save message in imap mbox */
    if ((sentfolder!=NULL) && (strlen(sentfolder)>0)) {
	r = do_append(do_getcurrentg(), sentfolder, g_waiting_doc->textview->a_tbuf);
	if (r) {
	    DLG_Alert(ALERT_warning, "Error saving message to sent folder");  
	    CON_WriteLine("Error writing message to sent folder");
	}
    }

    /* set answered flag xxx check errors*/
    if (g_waiting_doc->answer_mbox) {
	do_set_answered(do_getcurrentg(), g_waiting_doc->answer_mbox,g_waiting_doc->answer_num);
    }


    
    /* send msg smtp */
    TXT_SetBanner(g_waiting_doc->textview->a_tbuf, doc_submit_message);
    
    CON_SetBusy ( TRUE );

    smtp_start(&smtp_conn);
    
    r = smtp_send(smtp_conn,(char *)config_getstring("Email",NULL),
		  g_waiting_doc->textview->a_tbuf);

    smtp_close(smtp_conn);

    CON_SetBusy ( FALSE );

    if (r) {
	DLG_Alert(ALERT_warning, "Error sending message!");
	return;
    }
  
    
  /* Done. what now? */


  if (g_waiting_doc->textview) {
    TEXT_BUF *tbuf = g_waiting_doc->textview->a_tbuf;
    if (TF_ISSET(tbuf->flags, TF_OFFDSK)) {
      g_waiting_doc->stream = FIO_OpenStream(tbuf->lname, OPEN_RT);
      if (g_waiting_doc->stream == (FILE *)NULL)
	DOC_HandleRFileError();
    }
  }

  if (TF_ISSET(g_waiting_doc->flags, DOC_F_VERBOSE)) {
      TF_CLEAR(g_waiting_doc->textview->a_tbuf->flags, TF_CHANGED);
      CON_Write(MSG_FileSavedIn);
      CON_WriteLine(g_waiting_doc->local_fname);
  }

  if (g_waiting_doc->menu_cmd) {
      DOC_DoMenuCommand(g_waiting_doc->menu_cmd);
  } else {
      TF_CLEAR(g_waiting_doc->textview->a_tbuf->flags, TF_LOCKED);
      g_waiting_doc = NULL;
  }
  RFile_initialize();

  TF_CLEAR(tview->a_tbuf->flags, TF_CHANGED);
  TF_CLEAR(tview->a_tbuf->flags, TF_LOCKED);
  RFile_remove();
  TXT_SetBanner(tview->a_tbuf, MSG_SubmissionDone);
  if (config_getswitch("CloseOnSend",0)) TXT_Close(tview->a_tbuf->awp);
  g_waiting_doc = NULL;

}

void DOC_DoMenuCommand(int mid)
{
  TEXT_VIEW *atl = g_waiting_doc->textview;

  switch (mid) {
  case M_SEND:
      DOC_VerifyMsg(atl);
      DOC_SubmitMsg(atl);
      break;
  case M_VERIFY:
      DOC_VerifyMsg(atl);
      break;
  default:
    return;
  }
}
 
void DOC_ReplaceFile(char *fname,long fsize)
{
  Boolean reset_save = !RFile_error(TRUE);
  TEXT_VIEW *tview = g_waiting_doc->textview;
  int mid = g_waiting_doc->menu_cmd;

  if (mid == M_SEND && reset_save) {
    TF_CLEAR(tview->a_tbuf->flags, TF_CHANGED);
    TF_CLEAR(tview->a_tbuf->flags, TF_LOCKED);
    RFile_remove();
    TXT_SetBanner(tview->a_tbuf, MSG_SubmissionDone);
    if (config_getswitch("CloseOnSend",0)) TXT_Close(tview->a_tbuf->awp);
    g_waiting_doc = NULL;
  }
  else {
    if (DOC_Initialize(DOC_OP_REMOTE, fname, (long)fsize) == 0) {
      TF_SET(g_waiting_doc->flags, DOC_F_REMOVE);
      TF_SET_CHANGED(g_waiting_doc->textview->a_tbuf->flags);
      g_waiting_doc->save_tview = tview;
      tview = g_waiting_doc->textview;
    } else
      g_waiting_doc->textview = tview; /* restore it for clean up */

    if (mid == M_SEND)
	TXT_SetBanner(tview->a_tbuf, MSG_SubmissionFailed);
    else
      TXT_SetBanner(tview->a_tbuf,
		    (reset_save ? MSG_VerifyDone : MSG_VerifyFailed));
    if (g_waiting_doc->save_tview == (TEXT_VIEW_pt)NULL)
      DOC_ReInitialize(FALSE);
  }
}

void swap_and_release_tview(void)
{
  /*
   * g_waiting_doc->textview: filled up, ready, but not yet displayed.
   * g_waiting_doc->save_tview: The one to be replaced.
   * replace the file in save_tview by the one in textview.
   * Put in the new size of the file.  Also, swap the allocated buffer.
   */

  TEXT_BUF tbuf, *vf, *vt;

  vf = g_waiting_doc->textview->a_tbuf;
  vt = g_waiting_doc->save_tview->a_tbuf;

  bcopy(vt, &tbuf, sizeof(TEXT_BUF));  /* make a copy here */

  vt->flags = vf->flags;
  vt->buf   = vf->buf;
  vt->bsize = vf->bsize;
  vt->fsize = vf->fsize;
  vt->eop_pos = 0;
  vt->bop_pos = 0;
  vt->char_pos = 0;
  vt->mark_start = -1;
  vt->mark_end = -1;
  strcpy(vt->lname, vf->lname);
  strcpy(vt->ban_text, vf->ban_text);
  vt->stream = vf->stream;
  TXT_SetBanner(vt, vf->ban_text);

  vf->flags = tbuf.flags;
  vf->buf = tbuf.buf;
  vf->stream = tbuf.stream;
  strcpy(vf->lname, tbuf.lname);

  TF_CLEAR(vf->flags, TF_CHANGED);
  TF_CLEAR(vf->flags, TF_LOCKED);
  TXT_Close(g_waiting_doc->textview->a_tbuf->awp);
  g_waiting_doc->textview = g_waiting_doc->save_tview;
  g_waiting_doc->save_tview = (TEXT_VIEW *)NULL;
}

void DOC_ReInitialize(int net_is_dead)
{
  static Boolean doc_startup = TRUE;

  if (doc_startup) {
    g_waiting_doc = (DOCUMENT *)NULL;
    bzero(&g_document, sizeof(DOCUMENT));
    doc_startup = FALSE;
  } else if (g_waiting_doc) {
    if (g_waiting_doc->textview)
      TF_CLEAR(g_waiting_doc->textview->a_tbuf->flags, TF_LOCKED);
    if (g_waiting_doc->save_tview) {
      TF_CLEAR(g_waiting_doc->save_tview->a_tbuf->flags, TF_LOCKED);
      if (g_waiting_doc->textview) {
	TF_CLEAR(g_waiting_doc->textview->a_tbuf->flags, TF_CHANGED);
	TXT_Close(g_waiting_doc->textview->a_tbuf->awp);
      }
      g_waiting_doc->textview = g_waiting_doc->save_tview;
      g_waiting_doc->save_tview = (TEXT_VIEW_pt)NULL;
    }
    if (g_waiting_doc->textview && net_is_dead)
      TXT_SetBanner(g_waiting_doc->textview->a_tbuf, doc_standard_banner);
    if (g_waiting_doc->stream) FIO_CloseStream(g_waiting_doc->stream);
    if (g_waiting_doc->buf) free(g_waiting_doc->buf);
    bzero(g_waiting_doc, sizeof(DOCUMENT));
    g_waiting_doc = (DOCUMENT_pt)NULL;
  }
}

void DOC_ComposeMail(int  aCmdNumber, int ListNumber, int cmdIndex,long Selected,char *aStr)
{
  if (DOC_IsBusy() == FALSE) {
    RFile_initialize();

    if (aCmdNumber == M_AFS_EDIT)
      RL_open(ListNumber, (long)(Selected * -1),NULL);
    else
      RL_do_command(ListNumber, cmdIndex, (long)Selected, aStr);

    switch(aCmdNumber) {
    case M_QUICK_EDIT:
    case M_AFS_EDIT:
    case M_AFS_NEW:
      g_document.menu_cmd = aCmdNumber;
      TF_SET(g_document.flags, DOC_F_EDIT);
      break;
    case M_RESTORE:
    case M_REPLY_SENDER:
    case M_REPLY_READERS:
    case M_REPLY_ALL:
    case M_REPLY_DRAFT:
      TF_SET(g_document.flags, DOC_F_EDITBODY);
    case M_POST:
    case M_REPLY_FORWARD:
      TF_SET(g_document.flags, DOC_F_EDIT);
      TF_SET(g_document.flags, DOC_F_REMOVE);
      break;
    default:
      TF_CLEAR(g_document.flags, DOC_F_EDIT);
    }
  }
}

/*
 * remote data handling routines
 */

void DOC_FileArrived(char *fname,long  fsize)
{
  if (g_waiting_doc->menu_cmd == M_APND_REMOTE) {
    long total_size = g_waiting_doc->textview->a_tbuf->fsize + fsize;

    if (total_size >= PAGEABLE_BUFSIZE && g_waiting_doc->stream == (FILE *)0) {
      DLG_Alert(1, MSG_TooBigToAppend);
      DOC_HandleRFileError();
    } else {
      if (TXT_ReallocTview(g_waiting_doc->textview, (long)total_size)) {
	CON_WriteLine(MSG_NotEnoughMemory);
	DOC_HandleRFileError();
      } else {
	g_waiting_doc->fsize = fsize;             /* Store the size */
	RFile_read((long)g_waiting_doc->bufsize); /* Initiate 1st read */
      }
    }
  } else
    DOC_ReplaceFile(fname, (long)fsize);
}

void DOC_PickupFile(void)
{
  /*
   * The server has asked us to pick receive a file
   */

  char *fname = RFile_filename(); /* get the filename */
  long  fsize = RFile_datasize(); /* and the size */

  if (g_waiting_doc && g_waiting_doc->menu_cmd)
    DOC_FileArrived(fname, (long)fsize);
  else
    DOC_Initialize(DOC_OP_REMOTE, fname, (long)fsize);
}

void DOC_RemoteFileCreated(char *filename)
{
  /*
   * The server created the requested file.
   */

  strcpy(g_waiting_doc->local_fname, filename); /* save the name */

  if (g_waiting_doc->textview) {
    TEXT_BUF *tbuf = g_waiting_doc->textview->a_tbuf;
    if (TF_ISSET(tbuf->flags, TF_OFFDSK)) {
      g_waiting_doc->stream = FIO_OpenStream(tbuf->lname, OPEN_RT);
      if (g_waiting_doc->stream == (FILE *)NULL)
	DOC_HandleRFileError();
    }
  }
  DOC_RemoteWriteDone(); /* initiate the first write */
}

int DOC_SaveAsRemote(TEXT_VIEW *tview,char *filename)
{
  /*
   * Save the view on remote
   */

  if (DOC_IsBusy()) return(FALSE);

  RFile_initialize();
  /*  bzero(&g_document, sizeof(DOCUMENT)); */
  
  g_document.textview = tview;
  g_document.fsize = (long)(tview->a_tbuf->fsize);
  g_document.bufsize = RW_BUFSIZE;

  if (filename && *filename) TF_SET(g_document.flags, DOC_F_VERBOSE);
  else TF_SET(g_document.flags, DOC_F_REMOVE);

  TF_SET_LOCKED(tview->a_tbuf->flags); /* lock the view against changes */

  g_waiting_doc = &g_document;
  
  RFile_new(filename);
  return(TRUE);
}

void DOC_RemoteDataArrived(char *buf,long size)
{
  g_waiting_doc->buf[size] = 0;

  DOC_WriteToView(g_waiting_doc, g_waiting_doc->buf, (int)size);

  if (RFile_EOF()) {
    TEXT_BUF *atbuf;
    if (g_waiting_doc->buf) free(g_waiting_doc->buf);
    if (TF_ISSET(g_waiting_doc->flags, DOC_F_REMOVE)) RFile_remove();
    if (g_waiting_doc->save_tview) swap_and_release_tview();
    atbuf = g_waiting_doc->textview->a_tbuf;
    if (g_waiting_doc->menu_cmd == M_APND_REMOTE) {
      TF_SET_CHANGED(atbuf->flags);
      CON_WriteLine(MSG_AppendDone);
    }
    if (g_waiting_doc->stream) {
      FIO_CloseStream(g_waiting_doc->stream);
      atbuf->fsize = FIO_GetFileSize(g_waiting_doc->local_fname);
      TF_SET_OFFDSK(atbuf->flags);
      TF_SET_FLOAD(atbuf->flags);
    }
    DOC_ActivateView(atbuf);
    g_waiting_doc = (DOCUMENT_pt)NULL;
  } else {
    RFile_read((long)g_document.bufsize);
  }
}

void DOC_HandleRFileError(void)
{
  if (g_waiting_doc == NULL) return;

  if (g_waiting_doc->stream) FIO_CloseStream(g_waiting_doc->stream);

  if (g_waiting_doc->menu_cmd == M_APND_REMOTE ||
      !TF_ISSET(g_waiting_doc->flags, DOC_F_EXPECT_ERR))
    {
      DLG_Alert(ALERT_warning, MSG_Remote_IOerr);
      if (g_waiting_doc->buf) free(g_waiting_doc->buf);
      g_waiting_doc = (DOCUMENT_pt)NULL;
    }
/*
  else must be compose, we will receive a filename!
*/
}

void DOC_RemoteWriteDone(void)
{
  if (g_waiting_doc->offset >= g_waiting_doc->fsize) {
    if (TF_ISSET(g_waiting_doc->flags, DOC_F_VERBOSE)) {
      TF_CLEAR(g_waiting_doc->textview->a_tbuf->flags, TF_CHANGED);
      CON_Write(MSG_FileSavedIn);
      CON_WriteLine(g_waiting_doc->local_fname);
    }
    if (g_waiting_doc->menu_cmd) DOC_DoMenuCommand(g_waiting_doc->menu_cmd);
    else {
      TF_CLEAR(g_waiting_doc->textview->a_tbuf->flags, TF_LOCKED);
      g_waiting_doc = NULL;
    }
    RFile_initialize();
  } else {
    char buf[2048];
    int  bytes_read, len = sizeof(buf) - 1;
    if (g_waiting_doc->stream)
      bytes_read = FIO_ReadStream(g_waiting_doc->stream, buf, len);
    else
      bytes_read = TXT_GetData(g_waiting_doc->textview->a_tbuf,
			       buf, len, (long)(g_waiting_doc->offset));
    if (bytes_read > 0) {
      RFile_write(buf, (long)bytes_read);
/*    g_waiting_doc->offset += bytes_read;  */
      g_waiting_doc->offset += len;
    } else {
      g_waiting_doc->offset = g_waiting_doc->fsize; /* check for errors */
      DOC_RemoteWriteDone(); /* fake it so we get cleaned up */
    }
  }
}

/* JCS - Display the headers on this message? */
void DOC_UpdateHeaderDisplay(void)
{
    int num;
    int i;
    char line[100];
    char * headers;
    
    num = g_document.answer_num;
    
    if ( DOC_Initialize ( DOC_OP_LOCAL, NIL, 0 ) ) { 
	return;
    }
    
    if ( gShowFullHeaders ) { 
	headers = do_getheaders ( do_getcurrentg(), num );
	if ( headers ) { 
	    DOC_WriteToView ( &g_document, headers, strlen ( headers ) );    
	}
    }
    
    /* Put in full message body */
    i = 0;
    do {
	i++;
	RL_read(BODY_RL,i,line);
    } while (line[0]!='\0');
    
    for (;i<cur_bodyline;i++) {
	
	RL_read(BODY_RL,i,line);
	strcat(line,"\n");
	
	DOC_WriteToView(&g_document, line, strlen(line));  
    }
    
}
