/* SASL Config file API
 * Tim Martin (originally in Cyrus distribution)
 * $Id: ezconfig.c,v 1.7 2002/01/04 08:57:55 tmartin Exp $
 */
/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * Current Valid keys:
 *
 */


#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>

#include "xmalloc.h"
#include "ezconfig.h"

#define CONFIGNAME ".ezimail.ini"
#define GLOBALCONF (CONFIG_PREFIX "/etc/ezimail.defaults")

static char *
getfilename(void)
{
    char *homedir = getenv("HOME");
    char *ret;

    ret = xmalloc((MAXPATHLEN + 1) * sizeof(char *));

    if (homedir) {
	snprintf(ret, MAXPATHLEN, "%s/%s", homedir, CONFIGNAME);
    } else {
	snprintf(ret, MAXPATHLEN, "%s", CONFIGNAME);
    }

    return ret;
}

#define VL(x) 

struct configlist {
    char *key;
    char *value;
};

static struct configlist *configlist;
static int nconfiglist;
static int alloced = 0;

#define CONFIGLISTGROWSIZE 10 /* 100 */

int config_init(void)
{
    FILE *infile;
    int lineno = 0;
    char *filename;

    char buf[4096];
    char *p, *key;

    nconfiglist=0;

    filename = getfilename();

    infile = fopen(filename, "r");
    free(filename);
    if (!infile) {
	infile = fopen(GLOBALCONF, "r");
	if (!infile) {
	    VL(("Error opening %s",filename));
	    return 0;
	}
    }
    
    while (fgets(buf, sizeof(buf), infile)) {
	lineno++;

	VL(("reading config file lineno=%i\n",lineno));

	if (buf[strlen(buf)-1] == '\n') buf[strlen(buf)-1] = '\0';
	for (p = buf; *p && isspace((int) *p); p++);
	if (!*p || *p == '#') continue;

	key = p;
	while (*p && (isalnum((int) *p) || *p == '-' || *p == '_')) {
	    if (isupper((int) *p)) *p = tolower(*p);
	    p++;
	}
	if (*p != ':') {
	  return -1;
	}
	*p++ = '\0';

	while (*p && isspace((int) *p)) p++;
	
	if (!*p) {
	    /* empty value */
	    continue;
	}

	if (nconfiglist == alloced) {
	    alloced += CONFIGLISTGROWSIZE;
	    configlist=xrealloc((char *)configlist, alloced*sizeof(struct configlist));
	    if (configlist==NULL) return -1;
	}


	configlist[nconfiglist].key = xstrdup(key);
	configlist[nconfiglist].value = xstrdup(p);

	nconfiglist++;
    }
    fclose(infile);

    return 0;
}

const char *config_getstring(const char *key,const char *def)
{
    int opt;

    for (opt = 0; opt < nconfiglist; opt++) {
	if (!strcasecmp(key, configlist[opt].key))
	  return configlist[opt].value;
    }
    return def;
}

void config_setstring(const char *key, char *def)
{
    int opt;

    for (opt = 0; opt < nconfiglist; opt++) {
	if (!strcasecmp(key, configlist[opt].key))
	{
	    configlist[opt].value = xstrdup(def);
	    return;
	}
    }

    if (nconfiglist == alloced) {
	alloced += CONFIGLISTGROWSIZE;
	configlist=xrealloc((char *)configlist, alloced*sizeof(struct configlist));
	
    }

    configlist[nconfiglist].key = xstrdup(key);
    configlist[nconfiglist].value = xstrdup(def);
    
    nconfiglist++;
	
}

int config_getint(const char *key,int def)
{
    const char *val = config_getstring(key, (char *)0);

    if (!val) return def;
    if (!isdigit((int) *val) && (*val != '-' || !isdigit((int) val[1]))) return def;
    return atoi(val);
}

int config_getswitch(const char *key,int def)
{
    const char *val = config_getstring(key, (char *)0);

    if (!val) return def;

    if (*val == '0' || *val == 'n' ||
	(*val == 'o' && val[1] == 'f') || *val == 'f') {
	return 0;
    }
    else if (*val == '1' || *val == 'y' ||
	     (*val == 'o' && val[1] == 'n') || *val == 't') {
	return 1;
    }
    return def;
}

void config_setswitch(const char *key, int val)
{
    char *value;

    if (val == 0) 
	value = "false";
    else
	value = "true";
    
    config_setstring(key,value);
}


int config_getpos(char *name, config_position_t *pos,
		  int defx, int defy, int defdimx, int defdimy)
{
    /* xxx */
    pos->x = defx;
    pos->y = defy;
    pos->dimx = defdimx;
    pos->dimy = defdimy;

    return 0;
}



int config_save(void)
{
    FILE *stream;
    char *filename = getfilename();
    int opt;

    stream = fopen(filename,"w");
    free(filename);

    for (opt = 0; opt < nconfiglist; opt++) {

	/* only write keys that have values */
	if ( strlen ( configlist[opt].key ) > 0 && 
	     strlen ( configlist[opt].value ) > 0 ) {
	    fprintf(stream,"%s: %s\n",configlist[opt].key,configlist[opt].value);
	}
    }

    fprintf(stream,"\n");

    fclose(stream);

    return 0;
}
